@@ -6,8 +6,10 @@ use pyo3::pyclass::CompareOp;
 use pyo3::types::PyTuple;
 use pyo3::types::{PyDate, PyDateTime, PyDelta, PyDeltaAccess, PyDict, PyTime, PyTzInfo};
 use pyo3::IntoPyObjectExt;
-use speedate::MicrosecondsPrecisionOverflowBehavior;
-use speedate::{Date, DateTime, Duration, ParseError, Time, TimeConfig};
+use speedate::DateConfig;
+use speedate::{
+    Date, DateTime, DateTimeConfig, Duration, MicrosecondsPrecisionOverflowBehavior, ParseError, Time, TimeConfig,
+};
 use std::borrow::Cow;
 use std::collections::hash_map::DefaultHasher;
 use std::fmt::Write;
@@ -20,6 +22,7 @@ use super::Input;
 use crate::errors::ToErrorValue;
 use crate::errors::{ErrorType, ValError, ValResult};
 use crate::tools::py_err;
+use crate::validators::TemporalUnitMode;
 
 #[cfg_attr(debug_assertions, derive(Debug))]
 pub enum EitherDate<'py> {
@@ -49,7 +52,7 @@ pub fn pydate_as_date(py_date: &Bound<'_, PyAny>) -> PyResult<Date> {
 }
 
 impl<'py> EitherDate<'py> {
-    pub fn try_into_py(self, py: Python<'py>, input: &(impl Input<'py> + ?Sized)) -> ValResult<PyObject> {
+    pub fn try_into_py(self, py: Python<'py>, input: &(impl Input<'py> + ?Sized)) -> ValResult<Py<PyAny>> {
         match self {
             Self::Raw(date) => {
                 if date.year == 0 {
@@ -70,7 +73,7 @@ impl<'py> EitherDate<'py> {
 
     pub fn as_raw(&self) -> PyResult<Date> {
         match self {
-            Self::Raw(date) => Ok(date.clone()),
+            Self::Raw(date) => Ok(*date),
             Self::Py(py_date) => pydate_as_date(py_date),
         }
     }
@@ -110,8 +113,7 @@ impl<'py> IntoPyObject<'py> for EitherTimedelta<'py> {
     fn into_pyobject(self, py: Python<'py>) -> PyResult<Self::Output> {
         match self {
             Self::Raw(duration) => duration_as_pytimedelta(py, &duration),
-            Self::PyExact(py_timedelta) => Ok(py_timedelta),
-            Self::PySubclass(py_timedelta) => Ok(py_timedelta),
+            Self::PyExact(py_timedelta) | Self::PySubclass(py_timedelta) => Ok(py_timedelta),
         }
     }
 }
@@ -130,6 +132,93 @@ impl EitherTimedelta<'_> {
             Self::PySubclass(py_timedelta) => pytimedelta_subclass_as_duration(py_timedelta),
         }
     }
+
+    pub fn total_seconds(&self) -> PyResult<f64> {
+        match self {
+            Self::Raw(timedelta) => {
+                let mut days: i64 = i64::from(timedelta.day);
+                let mut seconds: i64 = i64::from(timedelta.second);
+                let mut microseconds = i64::from(timedelta.microsecond);
+                if !timedelta.positive {
+                    days = -days;
+                    seconds = -seconds;
+                    microseconds = -microseconds;
+                }
+
+                let days_seconds = (86_400 * days) + seconds;
+                if let Some(days_seconds_as_micros) = days_seconds.checked_mul(1_000_000) {
+                    let total_microseconds = days_seconds_as_micros + microseconds;
+                    Ok(total_microseconds as f64 / 1_000_000.0)
+                } else {
+                    // Fall back to floating-point operations if the multiplication overflows
+                    let total_seconds = days_seconds as f64 + microseconds as f64 / 1_000_000.0;
+                    Ok(total_seconds)
+                }
+            }
+            Self::PyExact(py_timedelta) => {
+                let days: i64 = py_timedelta.get_days().into(); // -999999999 to 999999999
+                let seconds: i64 = py_timedelta.get_seconds().into(); // 0 through 86399
+                let microseconds = py_timedelta.get_microseconds(); // 0 through 999999
+                let days_seconds = (86_400 * days) + seconds;
+                if let Some(days_seconds_as_micros) = days_seconds.checked_mul(1_000_000) {
+                    let total_microseconds = days_seconds_as_micros + i64::from(microseconds);
+                    Ok(total_microseconds as f64 / 1_000_000.0)
+                } else {
+                    // Fall back to floating-point operations if the multiplication overflows
+                    let total_seconds = days_seconds as f64 + f64::from(microseconds) / 1_000_000.0;
+                    Ok(total_seconds)
+                }
+            }
+            Self::PySubclass(py_timedelta) => py_timedelta
+                .call_method0(intern!(py_timedelta.py(), "total_seconds"))?
+                .extract(),
+        }
+    }
+
+    pub fn total_milliseconds(&self) -> PyResult<f64> {
+        match self {
+            Self::Raw(timedelta) => {
+                let mut days: i64 = i64::from(timedelta.day);
+                let mut seconds: i64 = i64::from(timedelta.second);
+                let mut microseconds = i64::from(timedelta.microsecond);
+                if !timedelta.positive {
+                    days = -days;
+                    seconds = -seconds;
+                    microseconds = -microseconds;
+                }
+
+                let days_seconds = (86_400 * days) + seconds;
+                if let Some(days_seconds_as_micros) = days_seconds.checked_mul(1_000_000) {
+                    let total_microseconds = days_seconds_as_micros + microseconds;
+                    Ok(total_microseconds as f64 / 1_000.0)
+                } else {
+                    // Fall back to floating-point operations if the multiplication overflows
+                    let total_seconds = days_seconds as f64 + microseconds as f64 / 1_000.0;
+                    Ok(total_seconds)
+                }
+            }
+            Self::PyExact(py_timedelta) => {
+                let days: i64 = py_timedelta.get_days().into(); // -999999999 to 999999999
+                let seconds: i64 = py_timedelta.get_seconds().into(); // 0 through 86399
+                let microseconds = py_timedelta.get_microseconds(); // 0 through 999999
+                let days_seconds = (86_400 * days) + seconds;
+                if let Some(days_seconds_as_micros) = days_seconds.checked_mul(1_000_000) {
+                    let total_microseconds = days_seconds_as_micros + i64::from(microseconds);
+                    Ok(total_microseconds as f64 / 1_000.0)
+                } else {
+                    // Fall back to floating-point operations if the multiplication overflows
+                    let total_milliseconds = days_seconds as f64 * 1_000.0 + f64::from(microseconds) / 1_000.0;
+                    Ok(total_milliseconds)
+                }
+            }
+            Self::PySubclass(py_timedelta) => {
+                let total_seconds: f64 = py_timedelta
+                    .call_method0(intern!(py_timedelta.py(), "total_seconds"))?
+                    .extract()?;
+                Ok(total_seconds / 1000.0)
+            }
+        }
+    }
 }
 
 impl<'py> TryFrom<&'_ Bound<'py, PyAny>> for EitherTimedelta<'py> {
@@ -244,7 +333,7 @@ impl<'py> IntoPyObject<'py> for EitherTime<'py> {
 impl EitherTime<'_> {
     pub fn as_raw(&self) -> PyResult<Time> {
         match self {
-            Self::Raw(time) => Ok(time.clone()),
+            Self::Raw(time) => Ok(*time),
             Self::Py(py_time) => pytime_as_time(py_time, None),
         }
     }
@@ -286,7 +375,7 @@ pub fn pydatetime_as_datetime(py_dt: &Bound<'_, PyAny>) -> PyResult<DateTime> {
 }
 
 impl<'py> EitherDateTime<'py> {
-    pub fn try_into_py(self, py: Python<'py>, input: &(impl Input<'py> + ?Sized)) -> ValResult<PyObject> {
+    pub fn try_into_py(self, py: Python<'py>, input: &(impl Input<'py> + ?Sized)) -> ValResult<Py<PyAny>> {
         match self {
             Self::Raw(dt) => {
                 if dt.date.year == 0 {
@@ -317,14 +406,18 @@ impl<'py> EitherDateTime<'py> {
 
     pub fn as_raw(&self) -> PyResult<DateTime> {
         match self {
-            Self::Raw(dt) => Ok(dt.clone()),
+            Self::Raw(dt) => Ok(*dt),
             Self::Py(py_dt) => pydatetime_as_datetime(py_dt),
         }
     }
 }
 
-pub fn bytes_as_date<'py>(input: &(impl Input<'py> + ?Sized), bytes: &[u8]) -> ValResult<EitherDate<'py>> {
-    match Date::parse_bytes(bytes) {
+pub fn bytes_as_date<'py>(
+    input: &(impl Input<'py> + ?Sized),
+    bytes: &[u8],
+    mode: TemporalUnitMode,
+) -> ValResult<EitherDate<'py>> {
+    match Date::parse_bytes_with_config(bytes, &DateConfig::builder().timestamp_unit(mode.into()).build()) {
         Ok(date) => Ok(date.into()),
         Err(err) => Err(ValError::new(
             ErrorType::DateParsing {
@@ -363,12 +456,16 @@ pub fn bytes_as_datetime<'py>(
     input: &(impl Input<'py> + ?Sized),
     bytes: &[u8],
     microseconds_overflow_behavior: MicrosecondsPrecisionOverflowBehavior,
+    mode: TemporalUnitMode,
 ) -> ValResult<EitherDateTime<'py>> {
     match DateTime::parse_bytes_with_config(
         bytes,
-        &TimeConfig {
-            microseconds_precision_overflow_behavior: microseconds_overflow_behavior,
-            unix_timestamp_offset: Some(0),
+        &DateTimeConfig {
+            time_config: TimeConfig {
+                microseconds_precision_overflow_behavior: microseconds_overflow_behavior,
+                unix_timestamp_offset: Some(0),
+            },
+            timestamp_unit: mode.into(),
         },
     ) {
         Ok(dt) => Ok(dt.into()),
@@ -386,13 +483,17 @@ pub fn int_as_datetime<'py>(
     input: &(impl Input<'py> + ?Sized),
     timestamp: i64,
     timestamp_microseconds: u32,
+    mode: TemporalUnitMode,
 ) -> ValResult<EitherDateTime<'py>> {
     match DateTime::from_timestamp_with_config(
         timestamp,
         timestamp_microseconds,
-        &TimeConfig {
-            unix_timestamp_offset: Some(0),
-            ..Default::default()
+        &DateTimeConfig {
+            time_config: TimeConfig {
+                unix_timestamp_offset: Some(0),
+                ..Default::default()
+            },
+            timestamp_unit: mode.into(),
         },
     ) {
         Ok(dt) => Ok(dt.into()),
@@ -420,12 +521,31 @@ macro_rules! nan_check {
     };
 }
 
-pub fn float_as_datetime<'py>(input: &(impl Input<'py> + ?Sized), timestamp: f64) -> ValResult<EitherDateTime<'py>> {
+pub fn float_as_datetime<'py>(
+    input: &(impl Input<'py> + ?Sized),
+    timestamp: f64,
+    mode: TemporalUnitMode,
+) -> ValResult<EitherDateTime<'py>> {
     nan_check!(input, timestamp, DatetimeParsing);
-    let microseconds = timestamp.fract().abs() * 1_000_000.0;
-    // checking for extra digits in microseconds is unreliable with large floats,
-    // so we just round to the nearest microsecond
-    int_as_datetime(input, timestamp.floor() as i64, microseconds.round() as u32)
+    match DateTime::from_float_with_config(
+        timestamp,
+        &DateTimeConfig {
+            time_config: TimeConfig {
+                unix_timestamp_offset: Some(0),
+                ..Default::default()
+            },
+            timestamp_unit: mode.into(),
+        },
+    ) {
+        Ok(dt) => Ok(dt.into()),
+        Err(err) => Err(ValError::new(
+            ErrorType::DatetimeParsing {
+                error: Cow::Borrowed(err.get_documentation().unwrap_or_default()),
+                context: None,
+            },
+            input,
+        )),
+    }
 }
 
 pub fn date_as_datetime<'py>(date: &Bound<'py, PyDate>) -> PyResult<EitherDateTime<'py>> {
@@ -549,6 +669,7 @@ pub struct TzInfo {
 #[pymethods]
 impl TzInfo {
     #[new]
+    #[pyo3(signature = (seconds = 0.0))]
     fn py_new(seconds: f32) -> PyResult<Self> {
         Self::try_from(seconds.trunc() as i32)
     }
@@ -574,7 +695,7 @@ impl TzInfo {
     }
 
     fn __repr__(&self) -> String {
-        format!("TzInfo({})", self.__str__())
+        format!("TzInfo({})", self.seconds)
     }
 
     fn __str__(&self) -> String {
